@@ -5,6 +5,10 @@ PACKAGE_TYPE="macosx10.9"
 if [[ "$PYTHONVERSION" =~ ^3.5 ]]; then
   PACKAGE_TYPE="macosx10.6"
 fi
+if [[ "$PYTHONVERSION" =~ ^3.10 ]]; then
+  PACKAGE_TYPE="macos11"
+fi
+
 
 # Handle prerelease versions
 PACKAGE_DIR="$PYTHONVERSION"
