@@ -12,12 +12,12 @@ import collections.abc
 import logging
 import uuid
 
-from uamqp import address, authentication, client, constants, errors, compat
-from uamqp.utils import get_running_loop
+from uamqp import address, authentication, client, constants, errors, compat, c_uamqp
 from uamqp.async_ops.connection_async import ConnectionAsync
 from uamqp.async_ops.receiver_async import MessageReceiverAsync
 from uamqp.async_ops.sender_async import MessageSenderAsync
 from uamqp.async_ops.session_async import SessionAsync
+from uamqp.async_ops.utils import get_dict_with_loop_if_needed
 
 try:
     TimeoutException = TimeoutError
@@ -38,13 +38,12 @@ class AMQPClientAsync(client.AMQPClient):
         - uamqp.authentication.SASLAnonymous
         - uamqp.authentication.SASLPlain
         - uamqp.authentication.SASTokenAsync
+        - uamqp.authentication.JWTTokenAsync
      If no authentication is supplied, SASLAnnoymous will be used by default.
     :type auth: ~uamqp.authentication.common.AMQPAuth
     :param client_name: The name for the client, also known as the Container ID.
      If no name is provided, a random GUID will be used.
     :type client_name: str or bytes
-    :param loop: A user specified event loop.
-    :type loop: ~asycnio.AbstractEventLoop
     :param debug: Whether to turn on network trace logs. If `True`, trace logs
      will be logged at INFO level. Default is `False`.
     :type debug: bool
@@ -105,14 +104,7 @@ class AMQPClientAsync(client.AMQPClient):
             keep_alive_interval=None,
             **kwargs):
 
-        if loop:
-            self.loop = loop
-        else:
-            try:
-                if not self.loop:  # from sub class instance
-                    self.loop = get_running_loop()
-            except AttributeError:
-                self.loop = get_running_loop()
+        self._internal_kwargs = get_dict_with_loop_if_needed(loop)
 
         super(AMQPClientAsync, self).__init__(
             remote_address,
@@ -146,9 +138,9 @@ class AMQPClientAsync(client.AMQPClient):
                     _logger.info("Keeping %r connection alive. %r",
                                  self.__class__.__name__,
                                  self._connection.container_id)
-                    await asyncio.shield(self._connection.work_async(), loop=self.loop)
+                    await asyncio.shield(self._connection.work_async(), **self._internal_kwargs)
                     start_time = current_time
-                await asyncio.sleep(1, loop=self.loop)
+                await asyncio.sleep(1, **self._internal_kwargs)
         except Exception as e:  # pylint: disable=broad-except
             _logger.info("Connection keep-alive for %r failed: %r.", self.__class__.__name__, e)
 
@@ -163,7 +155,7 @@ class AMQPClientAsync(client.AMQPClient):
 
     async def _client_run_async(self):
         """Perform a single Connection iteration."""
-        await asyncio.shield(self._connection.work_async(), loop=self.loop)
+        await asyncio.shield(self._connection.work_async(), **self._internal_kwargs)
 
     async def _redirect_async(self, redirect, auth):
         """Redirect the client endpoint using a Link DETACH redirect
@@ -177,7 +169,7 @@ class AMQPClientAsync(client.AMQPClient):
         # pylint: disable=protected-access
         if not self._connection._cbs:
             _logger.info("Closing non-CBS session.")
-            await asyncio.shield(self._session.destroy_async(), loop=self.loop)
+            await asyncio.shield(self._session.destroy_async(), **self._internal_kwargs)
         self._session = None
         self._auth = auth
         self._hostname = self._remote_address.hostname
@@ -197,8 +189,8 @@ class AMQPClientAsync(client.AMQPClient):
                     outgoing_window=self._outgoing_window,
                     handle_max=self._handle_max,
                     on_attach=self._on_attach,
-                    loop=self.loop),
-                loop=self.loop)
+                    **self._internal_kwargs),
+                **self._internal_kwargs)
             self._session = self._auth._session  # pylint: disable=protected-access
         elif self._connection._cbs:
             self._session = self._auth._session  # pylint: disable=protected-access
@@ -209,7 +201,11 @@ class AMQPClientAsync(client.AMQPClient):
                 outgoing_window=self._outgoing_window,
                 handle_max=self._handle_max,
                 on_attach=self._on_attach,
-                loop=self.loop)
+                **self._internal_kwargs)
+
+    @property
+    def loop(self):
+        return self._internal_kwargs.get("loop")
 
     async def open_async(self, connection=None):
         """Asynchronously open the client. The client can create a new Connection
@@ -242,10 +238,10 @@ class AMQPClientAsync(client.AMQPClient):
                 remote_idle_timeout_empty_frame_send_ratio=self._remote_idle_timeout_empty_frame_send_ratio,
                 error_policy=self._error_policy,
                 debug=self._debug_trace,
-                loop=self.loop)
+                **self._internal_kwargs)
             await self._build_session_async()
             if self._keep_alive_interval:
-                self._keep_alive_thread = asyncio.ensure_future(self._keep_alive_async(), loop=self.loop)
+                self._keep_alive_thread = asyncio.ensure_future(self._keep_alive_async(), **self._internal_kwargs)
         finally:
             if self._ext_connection:
                 connection.release_async()
@@ -267,13 +263,13 @@ class AMQPClientAsync(client.AMQPClient):
             return  # already closed.
         if not self._connection._cbs:  # pylint: disable=protected-access
             _logger.info("Closing non-CBS session.")
-            await asyncio.shield(self._session.destroy_async(), loop=self.loop)
+            await asyncio.shield(self._session.destroy_async(), **self._internal_kwargs)
         else:
             _logger.info("CBS session pending %r.", self._connection.container_id)
         self._session = None
         if not self._ext_connection:
             _logger.info("Closing exclusive connection %r.", self._connection.container_id)
-            await asyncio.shield(self._connection.destroy_async(), loop=self.loop)
+            await asyncio.shield(self._connection.destroy_async(), **self._internal_kwargs)
         else:
             _logger.info("Shared connection remaining open.")
         self._connection = None
@@ -314,7 +310,7 @@ class AMQPClientAsync(client.AMQPClient):
         :rtype: ~uamqp.message.Message
         """
         while not await self.auth_complete_async():
-            await asyncio.sleep(0.05, loop=self.loop)
+            await asyncio.sleep(0.05, **self._internal_kwargs)
         response = await asyncio.shield(
             self._session.mgmt_request_async(
                 message,
@@ -325,7 +321,7 @@ class AMQPClientAsync(client.AMQPClient):
                 encoding=self._encoding,
                 debug=self._debug_trace,
                 **kwargs),
-            loop=self.loop)
+            **self._internal_kwargs)
         return response
 
     async def auth_complete_async(self):
@@ -391,13 +387,12 @@ class SendClientAsync(client.SendClient, AMQPClientAsync):
         - uamqp.authentication.SASLAnonymous
         - uamqp.authentication.SASLPlain
         - uamqp.authentication.SASTokenAsync
+        - uamqp.authentication.JWTTokenAsync
      If no authentication is supplied, SASLAnnoymous will be used by default.
     :type auth: ~uamqp.authentication.common.AMQPAuth
     :param client_name: The name for the client, also known as the Container ID.
      If no name is provided, a random GUID will be used.
     :type client_name: str or bytes
-    :param loop: A user specified event loop.
-    :type loop: ~asycnio.AbstractEventLoop
     :param debug: Whether to turn on network trace logs. If `True`, trace logs
      will be logged at INFO level. Default is `False`.
     :type debug: bool
@@ -424,6 +419,11 @@ class SendClientAsync(client.SendClient, AMQPClientAsync):
      will assume successful receipt of the message and clear it from the queue. The
      default is `PeekLock`.
     :type receive_settle_mode: ~uamqp.constants.ReceiverSettleMode
+    :param desired_capabilities: The extension capabilities desired from the peer endpoint.
+     To create a desired_capabilities object, please do as follows:
+        - 1. Create an array of desired capability symbols: `capabilities_symbol_array = [types.AMQPSymbol(string)]`
+        - 2. Transform the array to AMQPValue object: `utils.data_factory(types.AMQPArray(capabilities_symbol_array))`
+    :type desired_capabilities: ~uamqp.c_uamqp.AMQPValue
     :param max_message_size: The maximum allowed message size negotiated for the Link.
     :type max_message_size: int
     :param link_properties: Metadata to be sent in the Link ATTACH frame.
@@ -469,7 +469,7 @@ class SendClientAsync(client.SendClient, AMQPClientAsync):
             error_policy=None,
             keep_alive_interval=None,
             **kwargs):
-        self.loop = loop or get_running_loop()
+        self._internal_kwargs = get_dict_with_loop_if_needed(loop)
         client.SendClient.__init__(
             self,
             target,
@@ -483,7 +483,7 @@ class SendClientAsync(client.SendClient, AMQPClientAsync):
 
         # AMQP object settings
         self.sender_type = MessageSenderAsync
-        self._pending_messages_lock = asyncio.Lock(loop=self.loop)
+        self._pending_messages_lock = asyncio.Lock(**self._internal_kwargs)
 
     async def _client_ready_async(self):
         """Determine whether the client is ready to start sending messages.
@@ -507,8 +507,9 @@ class SendClientAsync(client.SendClient, AMQPClientAsync):
                 properties=self._link_properties,
                 error_policy=self._error_policy,
                 encoding=self._encoding,
-                loop=self.loop)
-            await asyncio.shield(self.message_handler.open_async(), loop=self.loop)
+                desired_capabilities=self._desired_capabilities,
+                **self._internal_kwargs)
+            await asyncio.shield(self.message_handler.open_async(), **self._internal_kwargs)
             return False
         if self.message_handler.get_state() == constants.MessageSenderState.Error:
             raise errors.MessageHandlerError(
@@ -522,7 +523,8 @@ class SendClientAsync(client.SendClient, AMQPClientAsync):
     async def _transfer_message_async(self, message, timeout):
         sent = await asyncio.shield(
             self.message_handler.send_async(message, self._on_message_sent, timeout=timeout),
-            loop=self.loop)
+            **self._internal_kwargs
+            )
         if not sent:
             _logger.info("Message not sent, raising RuntimeError.")
             raise RuntimeError("Message sender failed to add message data to outgoing queue.")
@@ -561,6 +563,9 @@ class SendClientAsync(client.SendClient, AMQPClientAsync):
         """
         # pylint: disable=protected-access
         await self.message_handler.work_async()
+        await asyncio.shield(self._connection.work_async(), **self._internal_kwargs)
+        if self._connection._state == c_uamqp.ConnectionState.DISCARDING:
+            raise errors.ConnectionClose(constants.ErrorCodes.InternalServerError)
         self._waiting_messages = 0
         async with self._pending_messages_lock:
             self._pending_messages = await self._filter_pending_async()
@@ -568,7 +573,6 @@ class SendClientAsync(client.SendClient, AMQPClientAsync):
             _logger.info("Client told to backoff - sleeping for %r seconds", self._backoff)
             await self._connection.sleep_async(self._backoff)
             self._backoff = 0
-        await asyncio.shield(self._connection.work_async(), loop=self.loop)
         return True
 
     async def redirect_async(self, redirect, auth):
@@ -606,8 +610,9 @@ class SendClientAsync(client.SendClient, AMQPClientAsync):
         """Run the client asynchronously until all pending messages
         in the queue have been processed.
         """
-        while self.messages_pending():
-            await self.do_work_async()
+        running = True
+        while running and self.messages_pending():
+            running = await self.do_work_async()
 
     async def send_message_async(self, messages, close_on_done=False):
         """Send a single message or batched message asynchronously.
@@ -629,9 +634,10 @@ class SendClientAsync(client.SendClient, AMQPClientAsync):
                 self._pending_messages.append(message)
             pending_batch.append(message)
         await self.open_async()
+        running = True
         try:
-            while any([m for m in pending_batch if m.state not in constants.DONE_STATES]):
-                await self.do_work_async()
+            while running and any([m for m in pending_batch if m.state not in constants.DONE_STATES]):
+                running = await self.do_work_async()
             failed = [m for m in pending_batch if m.state == constants.MessageState.SendFailed]
             if any(failed):
                 details = {"total_messages": len(pending_batch), "number_failed": len(failed)}
@@ -642,7 +648,7 @@ class SendClientAsync(client.SendClient, AMQPClientAsync):
                     details['failed_messages'][failed_message] = exception
                 raise errors.ClientMessageError(exception, info=details)
         finally:
-            if close_on_done:
+            if close_on_done or not running:
                 await self.close_async()
 
     async def send_all_messages_async(self, close_on_done=True):
@@ -679,13 +685,12 @@ class ReceiveClientAsync(client.ReceiveClient, AMQPClientAsync):
         - uamqp.authentication.SASLAnonymous
         - uamqp.authentication.SASLPlain
         - uamqp.authentication.SASTokenAsync
+        - uamqp.authentication.JWTTokenAsync
      If no authentication is supplied, SASLAnnoymous will be used by default.
     :type auth: ~uamqp.authentication.common.AMQPAuth
     :param client_name: The name for the client, also known as the Container ID.
      If no name is provided, a random GUID will be used.
     :type client_name: str or bytes
-    :param loop: A user specified event loop.
-    :type loop: ~asycnio.AbstractEventLoop
     :param debug: Whether to turn on network trace logs. If `True`, trace logs
      will be logged at INFO level. Default is `False`.
     :type debug: bool
@@ -723,7 +728,7 @@ class ReceiveClientAsync(client.ReceiveClient, AMQPClientAsync):
      default is `PeekLock`.
     :type receive_settle_mode: ~uamqp.constants.ReceiverSettleMode
     :param desired_capabilities: The extension capabilities desired from the peer endpoint.
-     To create an desired_capabilities object, please do as follows:
+     To create a desired_capabilities object, please do as follows:
         - 1. Create an array of desired capability symbols: `capabilities_symbol_array = [types.AMQPSymbol(string)]`
         - 2. Transform the array to AMQPValue object: `utils.data_factory(types.AMQPArray(capabilities_symbol_array))`
     :type desired_capabilities: ~uamqp.c_uamqp.AMQPValue
@@ -774,7 +779,7 @@ class ReceiveClientAsync(client.ReceiveClient, AMQPClientAsync):
             error_policy=None,
             keep_alive_interval=None,
             **kwargs):
-        self.loop = loop or get_running_loop()
+        self._internal_kwargs = get_dict_with_loop_if_needed(loop)
         client.ReceiveClient.__init__(
             self,
             source,
@@ -815,8 +820,8 @@ class ReceiveClientAsync(client.ReceiveClient, AMQPClientAsync):
                 error_policy=self._error_policy,
                 encoding=self._encoding,
                 desired_capabilities=self._desired_capabilities,
-                loop=self.loop)
-            await asyncio.shield(self.message_handler.open_async(), loop=self.loop)
+                )
+            await asyncio.shield(self.message_handler.open_async(), **self._internal_kwargs)
             return False
         if self.message_handler.get_state() == constants.MessageReceiverState.Error:
             raise errors.MessageHandlerError(
@@ -842,7 +847,7 @@ class ReceiveClientAsync(client.ReceiveClient, AMQPClientAsync):
         now = self._counter.get_current_ms()
         if self._last_activity_timestamp and not self._was_message_received:
             # If no messages are coming through, back off a little to keep CPU use low.
-            await asyncio.sleep(0.05, loop=self.loop)
+            await asyncio.sleep(0.05, **self._internal_kwargs)
             if self._timeout > 0:
                 timespan = now - self._last_activity_timestamp
                 if timespan >= self._timeout:
@@ -999,7 +1004,7 @@ class ReceiveClientAsync(client.ReceiveClient, AMQPClientAsync):
 
 
 class AsyncMessageIter(collections.abc.AsyncIterator):  # pylint: disable=no-member
-    """Python 3.5 and 3.6 compatible asynchronous generator.
+    """Python 3.6 compatible asynchronous generator.
 
     :param recv_client: The receiving client.
     :type recv_client: ~uamqp.async_ops.client_async.ReceiveClientAsync
