@@ -29,6 +29,7 @@ class AMQPClient(object):
         - uamqp.authentication.SASLAnonymous
         - uamqp.authentication.SASLPlain
         - uamqp.authentication.SASTokenAuth
+        - uamqp.authentication.JWTTokenAuth
      If no authentication is supplied, SASLAnnoymous will be used by default.
     :type auth: ~uamqp.authentication.common.AMQPAuth
     :param client_name: The name for the client, also known as the Container ID.
@@ -259,6 +260,7 @@ class AMQPClient(object):
             self._build_session()
             if self._keep_alive_interval:
                 self._keep_alive_thread = threading.Thread(target=self._keep_alive)
+                self._keep_alive_thread.daemon = True
                 self._keep_alive_thread.start()
         finally:
             if self._ext_connection:
@@ -408,6 +410,7 @@ class SendClient(AMQPClient):
         - uamqp.authentication.SASLAnonymous
         - uamqp.authentication.SASLPlain
         - uamqp.authentication.SASTokenAuth
+        - uamqp.authentication.JWTTokenAuth
      If no authentication is supplied, SASLAnnoymous will be used by default.
     :type auth: ~uamqp.authentication.common.AMQPAuth
     :param client_name: The name for the client, also known as the Container ID.
@@ -439,6 +442,11 @@ class SendClient(AMQPClient):
      will assume successful receipt of the message and clear it from the queue. The
      default is `PeekLock`.
     :type receive_settle_mode: ~uamqp.constants.ReceiverSettleMode
+    :param desired_capabilities: The extension capabilities desired from the peer endpoint.
+     To create a desired_capabilities object, please do as follows:
+        - 1. Create an array of desired capability symbols: `capabilities_symbol_array = [types.AMQPSymbol(string)]`
+        - 2. Transform the array to AMQPValue object: `utils.data_factory(types.AMQPArray(capabilities_symbol_array))`
+    :type desired_capabilities: ~uamqp.c_uamqp.AMQPValue
     :param max_message_size: The maximum allowed message size negotiated for the Link.
     :type max_message_size: int
     :param link_properties: Metadata to be sent in the Link ATTACH frame.
@@ -521,7 +529,8 @@ class SendClient(AMQPClient):
                 link_credit=self._link_credit,
                 properties=self._link_properties,
                 error_policy=self._error_policy,
-                encoding=self._encoding)
+                encoding=self._encoding,
+                desired_capabilities=self._desired_capabilities)
             self.message_handler.open()
             return False
         if self.message_handler.get_state() == constants.MessageSenderState.Error:
@@ -641,13 +650,15 @@ class SendClient(AMQPClient):
         """
         # pylint: disable=protected-access
         self.message_handler.work()
+        self._connection.work()
+        if self._connection._state == c_uamqp.ConnectionState.DISCARDING:
+            raise errors.ConnectionClose(constants.ErrorCodes.InternalServerError)
         self._waiting_messages = 0
         self._pending_messages = self._filter_pending()
         if self._backoff and not self._waiting_messages:
             _logger.info("Client told to backoff - sleeping for %r seconds", self._backoff)
             self._connection.sleep(self._backoff)
             self._backoff = 0
-        self._connection.work()
         return True
 
     @property
@@ -790,6 +801,7 @@ class ReceiveClient(AMQPClient):
         - uamqp.authentication.SASLAnonymous
         - uamqp.authentication.SASLPlain
         - uamqp.authentication.SASTokenAuth
+        - uamqp.authentication.JWTTokenAuth
      If no authentication is supplied, SASLAnnoymous will be used by default.
     :type auth: ~uamqp.authentication.common.AMQPAuth
     :param client_name: The name for the client, also known as the Container ID.
@@ -832,7 +844,7 @@ class ReceiveClient(AMQPClient):
      default is `PeekLock`.
     :type receive_settle_mode: ~uamqp.constants.ReceiverSettleMode
     :param desired_capabilities: The extension capabilities desired from the peer endpoint.
-     To create an desired_capabilities object, please do as follows:
+     To create a desired_capabilities object, please do as follows:
         - 1. Create an array of desired capability symbols: `capabilities_symbol_array = [types.AMQPSymbol(string)]`
         - 2. Transform the array to AMQPValue object: `utils.data_factory(types.AMQPArray(capabilities_symbol_array))`
     :type desired_capabilities: ~uamqp.c_uamqp.AMQPValue
